% ip_sos.m: Stability conditions of PFMB observer-control system for the inverted pendulum.

clear all;
close all;

pdc = 1;

syms x1 x2 x3 x4 % x3 x4 are estimated states of x1 x2, respectively.
varX=[x1; x2; x3; x4];

%===== Parameters of Inverted Pendulum =====
g=9.8;
L=0.5;
mp=2;
Mc=8;
a=1/(mp+Mc);

x1min = -70*pi/180;
x1max = 70*pi/180;

xx1=x1min:0.01:x1max;

f1min = min(cos(xx1)./(4*L/3 - a*mp*L*cos(xx1).^2));
f1max = max(cos(xx1)./(4*L/3 - a*mp*L*cos(xx1).^2));

%===== Polynomial Model Parameters ===== (approximate)
% a1 = 1.7336; %Coefficient of polynomial function approximating tan(x1), tan(x1) = a1*x1 for x1min <= x1 <= x1max
% b1 = 0.8386; %Coefficient of polynomial function approximating sin(x1), sin(x1) = b1*x1 for x1min <= x1 <= x1max
% polyfit(xx1, tan(xx1), 1)
% polyfit(xx1, sin(xx1), 1)
a1 = 1.5534;
b1 = 0.8578;

A(:,:,1)=[0 1; f1min*(g*a1-a*mp*L*x2^2*b1) 0];
A(:,:,2)=[0 1; f1max*(g*a1-a*mp*L*x2^2*b1) 0];

B(:,:,1)=[0; -f1min*a];
B(:,:,2)=[0; -f1max*a];

Ah(:,:,1)=[0 1; f1min*(g*a1-a*mp*L*x4^2*b1) 0];
Ah(:,:,2)=[0 1; f1max*(g*a1-a*mp*L*x4^2*b1) 0];

Bh(:,:,1)=[0; -f1min*a];
Bh(:,:,2)=[0; -f1max*a];

% C(:,:,1)=[1 0];
% C(:,:,2)=[1 0];
% 
% Ch(:,:,1)=[1 0];
% Ch(:,:,2)=[1 0];

cc=1e-2;
C(:,:,1)=[1+cc*x2 0];
C(:,:,2)=[1+cc*x2 0];

Ch(:,:,1)=[1+cc*x4 0];
Ch(:,:,2)=[1+cc*x4 0];


% %===== Membership functions of fuzzy model =====
% f1 = cos(x1)./(4*L/3 - a*mp*L*cos(x1).^2);  % f1 = W1*f1min + W2*f1max
% W1 = (f1 - f1max)/(f1min - f1max);
% W2 = 1 - W1;

% Dimensions
[dimen_n, dimen_N, np]=size(A);
[trash1, dimen_m, trash2]=size(B);
[dimen_l, trash3, trash4]=size(C);

nc = np; % number of rules for controllers

% =============================================
% Initialize the sum of squares program
prog = sosprogram([varX]);

% Matrix variables
VECX = monomials(varX,[0]); % X
[prog,X] = sospolymatrixvar(prog,[VECX],[dimen_N dimen_N],'symmetric');

% Matrix variables
VECY = monomials(varX,[0]); % Y
[prog,Y] = sospolymatrixvar(prog,[VECY],[dimen_N dimen_N],'symmetric');

VECM = monomials([x4],[0:2:2]); % M
for k=1:nc
    [prog,M(:,:,k)] = sospolymatrixvar(prog,[VECM],[dimen_m dimen_N]);
end

VECN = monomials([x4],[0:1:1]); % N
for k=1:np
    [prog,N(:,:,k)] = sospolymatrixvar(prog,[VECN],[dimen_N dimen_l]);
end

% =============================================
% Predefined scalar polynomials for positive definite rather than semi-positive
varsigma1=1e-3;
varsigma2=1e-3;
varsigma3=1e-7;

alpha1 = 1e3;
alpha2 = 1e6;
beta = 1e-2;

% =============================================
%SOS conditions
expr = X - varsigma1*eye(dimen_N);
prog = sosmatrixineq(prog,expr,'Mineq');

expr = Y - varsigma2*eye(dimen_N);
prog = sosmatrixineq(prog,expr,'Mineq');

for k=1:nc
    expr = [1e7*eye(dimen_n) transpose(M(:,:,k)); M(:,:,k) 1];
    prog = sosmatrixineq(prog,expr,'Mineq');
end

Z = zeros(dimen_N,dimen_N);
I = eye(dimen_N);

switch(pdc)
    case(1) % PDC
        for i=1:np
            for j=1:np
                for k=1:np
                    if j<=k
                        expr = 0;
                        G11jk = Ah(:,:,j)*X - Bh(:,:,j)*M(:,:,k);
                        G11kj = Ah(:,:,k)*X - Bh(:,:,k)*M(:,:,j);
                        
                        G21jk = (A(:,:,i) - Ah(:,:,j))*X - (B(:,:,i) - Bh(:,:,j))*M(:,:,k);
                        G21kj = (A(:,:,i) - Ah(:,:,k))*X - (B(:,:,i) - Bh(:,:,k))*M(:,:,j);
                        
                        G12jk = N(:,:,j)*C(:,:,i);
                        G12kj = N(:,:,k)*C(:,:,i);
                        
                        G22jk = Y*A(:,:,i) - N(:,:,j)*C(:,:,i);
                        G22kj = Y*A(:,:,i) - N(:,:,k)*C(:,:,i);

                        Hjk = N(:,:,j)*(C(:,:,i) - Ch(:,:,k));
                        Hkj = N(:,:,k)*(C(:,:,i) - Ch(:,:,j));
                        
                        Xjk = [G11jk+transpose(G11jk),  transpose(G21jk),  I,  X;
                               G21jk,   -alpha2*I,  Z,  Z;
                               I,  Z,  -1/alpha1*Y,  Z;
                               X,  Z,  Z,  -1/beta*I];
                        
                        Xkj = [G11kj+transpose(G11kj),  transpose(G21kj),  I,  X;
                               G21kj,   -alpha2*I,  Z,  Z;
                               I,  Z,  -1/alpha1*Y,  Z;
                               X,  Z,  Z,  -1/beta*I];                    
                     
                        Yjk = [-alpha1*Y,  G12jk,  Z,  Hjk;
                               transpose(G12jk),  G22jk+transpose(G22jk),  Y,  -Hjk;
                               Z,  Y,  -1/alpha2*I,  Z;
                               transpose(Hjk),  -transpose(Hjk),  Z,  -beta*I];
                        
                        Ykj = [-alpha1*Y,  G12kj,  Z,  Hkj;
                               transpose(G12kj),  G22kj+transpose(G22kj),  Y,  -Hkj;
                               Z,  Y,  -1/alpha2*I,  Z;
                               transpose(Hkj),  -transpose(Hkj),  Z,  -beta*I];
                        
                        expr = - Xjk - Xkj - varsigma3*eye(4*dimen_N);
                        prog = sosmatrixineq(prog,expr,'Mineq');
                        expr = - Yjk - Ykj - varsigma3*eye(4*dimen_N);
                        prog = sosmatrixineq(prog,expr,'Mineq');
                    end
                end
            end
        end
end

% =============================================
% Call solver
[prog, info] = sossolve(prog);

% Solutions
SOLX = sosgetsol(prog,X);
SOLY = sosgetsol(prog,Y);

if det(SOLX)~=0 && det(SOLY)~=0
    for k=1:nc
        F(:,:,k) = sosgetsol(prog,M(:,:,k));
        F(:,:,k) = F(:,:,k)/SOLX;
    end
    F
    for k=1:np
        LL(:,:,k) = sosgetsol(prog,N(:,:,k));
        LL(:,:,k) = inv(SOLY)*LL(:,:,k);
    end
    LL
end

% Feasibility criterion
if abs(info.feasratio - 1)<0.5 && info.pinf==0 && info.dinf==0 && info.numerr<=1
    tmin = -1;
else
    tmin = 1;
end

% [max_meig,meig]=sosOBSvf(F,L);
% max_meig




