% ip_sossp.m: Stability conditions of PFMB observer-control system with sampled-output measurement for the inverted pendulum.

clear all;
close all;

pdc = 1;

syms x1 x2 x3 x4 % x3 x4 are estimated states of x1 x2, respectively.
varX=[x1; x2; x3; x4];

%===== Parameters of Inverted Pendulum =====
g=9.8;
L=0.5;
mp=2;
Mc=8;
a=1/(mp+Mc);

x1min = -70*pi/180;
x1max = 70*pi/180;

xx1=x1min:0.01:x1max;

f1min = min(cos(xx1)./(4*L/3 - a*mp*L*cos(xx1).^2));
f1max = max(cos(xx1)./(4*L/3 - a*mp*L*cos(xx1).^2));

%===== Polynomial Model Parameters ===== (approximate)
% a1 = 1.7336; %Coefficient of polynomial function approximating tan(x1), tan(x1) = a1*x1 for x1min <= x1 <= x1max
% b1 = 0.8386; %Coefficient of polynomial function approximating sin(x1), sin(x1) = b1*x1 for x1min <= x1 <= x1max
% polyfit(xx1, tan(xx1), 1)
% polyfit(xx1, sin(xx1), 1)
a1 = 1.5534;
b1 = 0.8578;

A(:,:,1)=[0 1; f1min*(g*a1-a*mp*L*x2^2*b1) 0];
A(:,:,2)=[0 1; f1max*(g*a1-a*mp*L*x2^2*b1) 0];

B(:,:,1)=[0; -f1min*a];
B(:,:,2)=[0; -f1max*a];

Ah(:,:,1)=[0 1; f1min*(g*a1-a*mp*L*x4^2*b1) 0];
Ah(:,:,2)=[0 1; f1max*(g*a1-a*mp*L*x4^2*b1) 0];

Bh(:,:,1)=[0; -f1min*a];
Bh(:,:,2)=[0; -f1max*a];

% C(:,:,1)=[1 0];
% C(:,:,2)=[1 0];

cc = 0.1;
C(:,:,1)=[1+cc 0];
C(:,:,2)=[1-cc 0];

% %===== Membership functions of fuzzy model =====
% f1 = cos(x1)./(4*L/3 - a*mp*L*cos(x1).^2);  % f1 = W1*f1min + W2*f1max
% W1 = (f1 - f1max)/(f1min - f1max);
% W2 = 1 - W1;

% Dimensions
[dimen_n, dimen_N, np]=size(A);
[trash1, dimen_m, trash2]=size(B);
[dimen_l, trash3, trash4]=size(C);

nc = np; % number of rules for controllers

% =============================================
% Initialize the sum of squares program
prog = sosprogram([varX]);

% Matrix variables
VECX = monomials(varX,[0]); % X
[prog,X] = sospolymatrixvar(prog,[VECX],[dimen_N dimen_N],'symmetric');

% Matrix variables
VECY = monomials(varX,[0]); % Y
[prog,Y] = sospolymatrixvar(prog,[VECY],[dimen_N dimen_N],'symmetric');

% Matrix variables
VECQ = monomials(varX,[0]); % Q
[prog,Q] = sospolymatrixvar(prog,[VECQ],[2*dimen_N 2*dimen_N],'symmetric');

VECM = monomials([x4],[0:2:2]); % M
for k=1:nc
    [prog,M(:,:,k)] = sospolymatrixvar(prog,[VECM],[dimen_m dimen_N]);
end

VECN = monomials([x4],[0:2:2]); % N
for k=1:np
    [prog,N(:,:,k)] = sospolymatrixvar(prog,[VECN],[dimen_N dimen_l]);
end


% =============================================
% Predefined scalar polynomials for positive definite rather than semi-positive
varsigma1=1e-3;
varsigma2=1e-3;
varsigma3=1e-7;

alpha1=1e6;
alpha2=1e5;
alpha3=1e3;
alpha4=1e3;
beta=1e-2;
gamma=1e-1;

h=0.05;

% =============================================
%SOS conditions
expr = X - varsigma1*eye(dimen_N);
prog = sosmatrixineq(prog,expr,'Mineq');

expr = Y - varsigma1*eye(dimen_N);
prog = sosmatrixineq(prog,expr,'Mineq');

expr = Q - varsigma2*eye(2*dimen_N);
prog = sosmatrixineq(prog,expr,'Mineq');

Z1 = zeros(dimen_N,dimen_N);
Z2 = zeros(2*dimen_N,2*dimen_N);
I1 = eye(dimen_N);
I2 = eye(2*dimen_N);

switch(pdc)
    case(1)
        for m=1:np
            for j=1:np
                for k=1:np
                    if j<=k
                        expr = 0;
                        G11jk = Ah(:,:,j)*X - Bh(:,:,j)*M(:,:,k);
                        G11kj = Ah(:,:,k)*X - Bh(:,:,k)*M(:,:,j);
                        
                        G21jk = (A(:,:,m) - Ah(:,:,j))*X - (B(:,:,m) - Bh(:,:,j))*M(:,:,k);
                        G21kj = (A(:,:,m) - Ah(:,:,k))*X - (B(:,:,m) - Bh(:,:,k))*M(:,:,j);                        
                        
                        P11jk = [G11jk,  Z1;
                            G21jk,  - alpha1/2*I1];
                        P11kj = [G11kj,  Z1;
                            G21kj,  - alpha1/2*I1];
                        
                        P21 = [Z1 Z1];
                        
                        P22 = -alpha4*I1;
                        
                        P31jk = h*[G11jk,  Z1;
                            G21jk,  Z1];
                        P31kj = h*[G11kj,  Z1;
                            G21kj,  Z1];
                        
                        P32 = [Z1; Z1];
                        
                        P33 = - [2*gamma*X Z1; Z1 alpha3*I1];
                        
                        P41 = Z2;
                        
                        P42 = [X;  Z1];
                        
                        P43 = [Z1  Z1;
                            I1  Z1];
                        
                        P44 = [- 1/beta*I1, Z1;
                            Z1, - 1/alpha2*Y];
                        
                        Xjk = [P11jk+transpose(P11jk)  transpose(P21)  transpose(P31jk)  transpose(P41);
                            P21  P22  transpose(P32)  transpose(P42);
                            P31jk  P32  P33  transpose(P43);
                            P41  P42  P43  P44];
                        
                        Xkj = [P11kj+transpose(P11kj)  transpose(P21)  transpose(P31kj)  transpose(P41);
                            P21  P22  transpose(P32)  transpose(P42);
                            P31kj  P32  P33  transpose(P43);
                            P41  P42  P43  P44];
                        
                        expr = - Xjk - Xkj - varsigma3*eye(7*dimen_N);
                        prog = sosmatrixineq(prog,expr,'Mineq');
                        
                    end
                end
            end
        end
        
        
        for i=1:np
            for l=1:np
                for m=1:np
                    for j=1:np
                        expr = 0;
                        G22jk = Y*A(:,:,m);
                        
                        Hjk = N(:,:,j)*(C(:,:,i) - C(:,:,l));
                        
                        Kjk = N(:,:,j)*C(:,:,i);                   
                        
                        P11jk = [Z1,  Z1;
                            Z1,  G22jk] - Q/2;
                        
                        P21jk = [Z1,  Z1;
                            transpose(Kjk),  - transpose(Kjk)] + Q;
                        
                        P22jk = - 2*Q;
                        
                        P31jk = h*[Z1,  Z1;
                            Z1,  G22jk];
                        
                        P32jk = h*[Z1,  Kjk;
                            Z1,  - Kjk];
                        
                        P33jk = gamma^2*Q - [alpha2*Y  Z1; Z1  2*gamma*Y];
                      
                        Pjk = [P11jk+transpose(P11jk),  transpose(P21jk),  Z2,  transpose(P31jk);
                            P21jk,  P22jk,  Q,  transpose(P32jk);
                            Z2,  Q,  -Q,  Z2;
                            P31jk,  P32jk,  Z2,  P33jk];
                        
                        Y12 = [Z1; Y; zeros(6*dimen_N,dimen_N)];
                        Y13 = [Z1; Z1; Y; zeros(5*dimen_N,dimen_N)];
                        Y14 = [zeros(7*dimen_N,dimen_N); Y];
                        Y15jk = [Hjk; -Hjk; zeros(4*dimen_N,dimen_N); h*Hjk; -h*Hjk];
                       
                        Yjk = [Pjk,  Y12,  Y13,  Y14,  Y15jk;
                            transpose(Y12),  -1/alpha1*I1,  Z1,  Z1,  Z1;
                            transpose(Y13),  Z1,  -1/alpha4*I1,  Z1,  Z1;
                            transpose(Y14),  Z1,  Z1,  -1/alpha3*I1,  Z1;
                            transpose(Y15jk),  Z1,  Z1,  Z1,  -beta*I1];
                        
                        expr = - Yjk - varsigma3*eye(12*dimen_N);
                        prog = sosmatrixineq(prog,expr,'Mineq');
                    end
                end
            end
        end
        
end

% =============================================
% Call solver
[prog, info] = sossolve(prog);

% Solutions
SOLX = sosgetsol(prog,X);
SOLY = sosgetsol(prog,Y);

if det(SOLX)~=0 && det(SOLY)~=0
    for k=1:nc
        F(:,:,k) = sosgetsol(prog,M(:,:,k));
        F(:,:,k) = F(:,:,k)/SOLX;
    end
    F
    for k=1:np
        LL(:,:,k) = sosgetsol(prog,N(:,:,k));
        LL(:,:,k) = inv(SOLY)*LL(:,:,k);
    end
    LL
end

% Feasibility criterion
if abs(info.feasratio - 1)<0.5 && info.pinf==0 && info.dinf==0 && info.numerr<=1
    tmin = -1;
else
    tmin = 1;
end

% [max_meig,meig]=sosOBSvf(F,L);
% max_meig




